import papermill as pm
import pytest

from pathlib import Path

path_to_exploratory = Path(__file__).parents[1] / 'exploratory'
path_to_exploratory_tests = Path(__file__).parents[1] / 'exploratory-tests'


@pytest.mark.parametrize(
    'name',
    [
        'example.ipynb',
        # other notebooks...
    ])
def test_exploratory(name):
    """
    Smoke tests for notebooks in exploratory/ (runs the noteboks but does not
    check output)
    """
    path_to_exploratory_tests.mkdir(exist_ok=True)

    pm.execute_notebook(path_to_exploratory / name,
                        path_to_exploratory_tests / name)



from ploomber.executors import Serial
from ploomber.spec import DAGSpec


def test_pipeline():
    """
    This is a smoke test, checking that the pipeline runs (but not the output)

    NOTE: it's common for pipelines to take hours to run, a way to make this
    test feasible is to run it here with a sample of the data and save results
    in a different folder to prevent overwriting your results.
    """
    # load dag
    dag = DAGSpec.find().to_dag()

    # change executor settings: you can use "pytest --pdb" to start a debugging
    # session if the test fails. Calling dag['task'].debug() is another
    # option
    dag.executor = Serial(build_in_subprocess=False, catch_exceptions=False)

    # a third approach for debugging is to use: import IPython; IPython.embed()
    # to start an interactive session at this point. To do so, you must call
    # "pytest -s"

    dag.build()



from pathlib import Path

import numpy as np
import pandas as pd


def get(product):
    df = pd.DataFrame(np.random.rand(100, 2), columns=['a', 'b'])
    df.loc[:10, 'a'] = np.nan
    Path(str(product)).parent.mkdir(parents=True, exist_ok=True)
    df.to_parquet(str(product))



from pathlib import Path

import pandas as pd


def add_column(upstream, product):
    df = pd.read_parquet(str(upstream['remove_nas']))
    df['c'] = df.a + df.b
    Path(str(product)).parent.mkdir(parents=True, exist_ok=True)
    df.to_parquet(str(product))



"""
API for creating scaffolding projects
"""
import itertools
import os
import shutil
import re
from pathlib import Path
import argparse

try:
    from importlib import resources
except ImportError:
    # python < 3.7
    import importlib_resources as resources

import ploomber_scaffold


def copy_template(path):
    """Copy template files to path
    """
    with resources.path(ploomber_scaffold, 'template') as path_to_template:
        shutil.copytree(path_to_template, path)


def is_valid_package_name(package_name):
    match = re.match(r'^[\w]+$', package_name) or False
    return match and not package_name[0].isnumeric()


def last_part(project_path):
    return project_path.parts[-1]


def render_template(path, package_name):
    """
    Replace references to "package_name" in the template for the name selected
    by the user
    """
    for dirpath, _, filenames in os.walk(path):
        paths = [
            Path(dirpath, f) for f in filenames if Path(f).suffix in
            {'.py', '.md', '.yml', '.yaml', '.sh', '.sql'} or f == '.gitignore'
        ]

        for p in itertools.chain(paths):
            p = Path(p)
            new = p.read_text().replace('package_name', package_name)
            p.write_text(new)

    pkg_dir = path / 'src' / 'package_name'
    pkg_dir.rename(path / 'src' / package_name)


def cli(project_path):
    project_path = None if not project_path else Path(project_path)

    print('Project names should be alphanumeric, all-lowercase.\n'
          'The first character cannot be numeric.\n')

    if project_path is None:
        project_path = Path(input('Project name: '))
        pkg_name = last_part(project_path)

        while not is_valid_package_name(pkg_name):
            print(f'{pkg_name!r} is not a valid name, choose another.')
            pkg_name = input('Project name: ')
    else:
        pkg_name = last_part(project_path)

        if not is_valid_package_name(pkg_name):
            raise ValueError('"%s" is not a valid package identifier, '
                             'choose another.' % pkg_name)

    if project_path.is_dir() and len(os.listdir(project_path)):
        raise ValueError(f'{str(project_path)!r} is a non-empty directory')
    elif project_path.is_file():
        raise ValueError(f'{str(project_path)!r} is an existing file')

    copy_template(project_path)
    render_template(project_path, pkg_name)
    readme_path = project_path / 'README.md'
    print(f'Done. Check out {readme_path!s} to get started.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Create scaffold')
    parser.add_argument('--path', type=str, help='Path', default=None)
    args = parser.parse_args()
    cli(project_path=args.path)


import io
import re
import ast
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

# read __version__ = VERSION from __init__.py
with open('src/package_name/__init__.py', 'rb') as f:
    VERSION = str(
        ast.literal_eval(
            _version_re.search(f.read().decode('utf-8')).group(1)))


def read(*names, **kwargs):
    return io.open(join(dirname(__file__), *names),
                   encoding=kwargs.get('encoding', 'utf8')).read()


# NOTE: keep these lists updated as you add more dependencies to your project
# if you rather have any dependency installed via conda, add it here and in
# the environment.yml file

# minimum dependencies for deployment
REQUIRES = [
    'pyarrow',
    'numpy',
    'pandas',
    'ploomber',
]

# extra dependencies for dewvelopment. e.g. run tests, build docs,
# train new models, generate exploratory notebooks, etc
REQUIRES_DEV = [
    'pytest',
    'nox',
    'pyyaml',
    'invoke',
    'flake8',
    'jupyter',
    'matplotlib',
]

setup(
    name='package_name',
    version=VERSION,
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    # Include any non-Python files with these extensions, for details:
    # https://setuptools.readthedocs.io/en/latest/userguide/datafiles.html
    package_data={"": ["*/*.sql", "*/*.ipynb", "notebooks/*.py"]},
    install_requires=REQUIRES,
    extras_require={
        'dev': REQUIRES_DEV,
    },
)



"""
nox configuration file.
docs: https://nox.thea.codes/en/stable/index.html
"""
import sys
from pathlib import Path

import nox


def env_file_name():
    """
    Get environment file to use, prioritizing environment.lock.yml
    """
    env_file = 'environment.lock.yml'

    if not Path(env_file).exists():
        print('[WARNING] environment.lock.yaml does not '
              'exist, looking for environment.yml...')

        env_file = 'environment.yml'

        if not Path(env_file).exists():
            print('[ERROR] No conda environment file found. Aborting.')
            sys.exit(1)
        else:
            print('Using environment.yml...')

    return env_file


env_file = env_file_name()


@nox.session(venv_backend='conda', python='3.8')
def tests(session):
    session._run('conda', 'env', 'update', '--prefix',
                 session.virtualenv.location, '--file', env_file)

    session.install('--editable', '.[dev]')

    session.run('pytest', 'tests/')

    print('Generating environment.lock.yml...')
    session.run('conda', 'env', 'export', '--prefix',
                session.virtualenv.location, '--file', 'environment.lock.yml')



from pathlib import Path

import pandas as pd


def remove_nas(upstream, product):
    df = pd.read_parquet(str(upstream['get']))
    df = df[~df.a.isna()]
    Path(str(product)).parent.mkdir(parents=True, exist_ok=True)
    df.to_parquet(str(product))



"""
Command line interface.

Requires invoke (pip install invoke). For help run "invoke -h", to list
commands "invoke -l" (you can also run the shorthand command "inv")

Source code for simple commands can be included here, for large ones, save it
in the lib/ folder and import it here.
"""
from invoke import task
from lib import conda, versioneer


@task
def setup(c):
    """Setup development environment
    """
    print('Creating conda environment...')
    c.run('conda env create environment.yml --force')
    print('Installing package...')
    conda.run_in_env(c, 'pip install --editable .[dev]', env='package_name')
    print('Done! Activate your environment with: conda activate package_name')
    print('Then run the pipeline with: ploomber build')


@task(
    help={
        'inplace':
        'Runs tests in the current environment '
        '(calling pytest directly). Does not generate lock file'
    })
def test(c, inplace=False):
    """Run tests and generates lock file
    """
    if inplace:
        print('Running tests in the current environment...')
        c.run('pytest tests/', pty=True)
    else:
        c.run('nox', pty=True)


@task
def version(c):
    """
    Create a new version
    1. Set new stable version in package_name/__init__.py
    2. Update header in CHANGELOG file, and ask to review CHANGELOG
    3. Create commit for new version, create git tag, and push
    4. Set new development version in package_name/__init__.py and CHANGELOG
    5. Commit new development version, and push

    (more details in CONTRIBUTING.md)
    """
    versioneer.version(project_root='.', tag=True)



"""
Managing project versions
"""
import sys
import ast
import re
import subprocess
import datetime
import os
from pathlib import Path


def replace_in_file(path_to_file, original, replacement):
    """Replace string in file
    """
    with open(path_to_file, 'r+') as f:
        content = f.read()
        updated = content.replace(original, replacement)
        f.seek(0)
        f.write(updated)
        f.truncate()


def call(*args, **kwargs):
    return subprocess.run(*args, **kwargs, check=True)


def _input(prompt):
    return input(prompt)


def input_str(prompt, default):
    separator = ' ' if len(prompt.splitlines()) == 1 else '\n'
    response = _input(prompt + f'{separator}(Default: {default}): ')

    if not response:
        response = default

    return response


def input_confirm(prompt, abort):
    separator = ' ' if len(prompt.splitlines()) == 1 else '\n'
    response_raw = _input(prompt + f'{separator}Confirm? [y/n]: ')
    response = response_raw in {'y', 'Y', 'yes'}

    if not response and abort:
        print('Aborted!')
        sys.exit(1)

    return response


class Versioner:
    """Utility functions to manage versions
    """
    def __init__(self, project_root='.'):
        path_to_src = Path(project_root, 'src')

        dirs = [
            f for f in os.listdir(path_to_src)
            if Path('src', f).is_dir() and not f.endswith('.egg-info')
        ]

        if len(dirs) != 1:
            raise ValueError(f'src/ must have a single folder, got: {dirs}')

        self.package_name = dirs[0]
        self.PACKAGE = path_to_src / self.package_name

        if Path(project_root, 'CHANGELOG.rst').exists():
            self.path_to_changelog = Path(project_root, 'CHANGELOG.rst')
        elif Path(project_root, 'CHANGELOG.md').exists():
            self.path_to_changelog = Path(project_root, 'CHANGELOG.md')
        else:
            self.path_to_changelog = None

    def current_version(self):
        """Returns the current version in __init__.py
        """
        _version_re = re.compile(r'__version__\s+=\s+(.*)')

        with open(self.PACKAGE / '__init__.py', 'rb') as f:
            VERSION = str(
                ast.literal_eval(
                    _version_re.search(f.read().decode('utf-8')).group(1)))

        return VERSION

    def release_version(self):
        """
        Returns a release version number
        e.g. 2.4.4dev -> v.2.2.4
        """
        current = self.current_version()

        if 'dev' not in current:
            raise ValueError('Current version is not a dev version')

        return current.replace('dev', '')

    def bump_up_version(self):
        """
        Gets gets a release version and returns a the next value value.
        e.g. 1.2.5 -> 1.2.6dev
        """
        # Get current version
        current = self.current_version()

        if 'dev' in current:
            raise ValueError('Current version is dev version, new dev '
                             'versions can only be made from release versions')

        # Get Z from X.Y.Z and sum 1
        tokens = current.split('.')

        # if just released a major version, add a 0 so we bump up a subversion
        # e.g. from 0.8 -> 0.8.0, then new dev version becomes 0.8.1dev
        if len(tokens) == 2:
            tokens.append('0')

        new_subversion = int(tokens[-1]) + 1

        # Replace new_subversion in current version
        tokens[-1] = str(new_subversion)
        new_version = '.'.join(tokens) + 'dev'

        return new_version

    def commit_version(self, new_version, msg_template, tag=False):
        """
        Replaces version in  __init__ and optionally creates a tag in the git
        repository (also saves a commit)
        """
        current = self.current_version()

        # replace new version in __init__.py
        replace_in_file(self.PACKAGE / '__init__.py', current, new_version)

        # Run git add and git status
        print('Adding new changes to the repository...')
        call(['git', 'add', '--all'])
        call(['git', 'status'])

        # Commit repo with updated dev version
        print('Creating new commit release version...')
        msg = msg_template.format(package_name=self.package_name,
                                  new_version=new_version)
        call(['git', 'commit', '-m', msg])

        # Create tag
        if tag:
            print('Creating tag {}...'.format(new_version))
            message = msg_template.format(package_name=self.package_name,
                                          new_version=new_version)
            call(['git', 'tag', '-a', new_version, '-m', message])

            print('Pushing tags...')
            call(['git', 'push', 'origin', new_version])

    def update_changelog_release(self, new_version):
        """Updates changelog file, adding a new section
        """
        current_version = self.current_version()

        # update CHANGELOG header
        header_current = make_header(current_version,
                                     self.path_to_changelog,
                                     add_date=False)

        header_new = make_header(new_version,
                                 self.path_to_changelog,
                                 add_date=True)

        replace_in_file(self.path_to_changelog, header_current, header_new)

    def add_changelog_new_dev_section(self, dev_version):
        if self.path_to_changelog:
            if self.path_to_changelog.suffix == '.rst':
                start_current = 'CHANGELOG\n========='
            else:
                start_current = '# CHANGELOG'

            new_header = make_header(dev_version, self.path_to_changelog)
            start_new = f'{start_current}\n\n{new_header}'
            replace_in_file(self.path_to_changelog, start_current, start_new)
        else:
            print('No CHANGELOG.{rst,md} found, skipping changelog editing...')


def make_header(content, path, add_date=False):
    if add_date:
        today = datetime.datetime.now().strftime('%Y-%m-%d')
        content += f' ({today})'

    if path.suffix == '.md':
        return f'## {content}'
    elif path.suffix == '.rst':
        return f'{content}\n' + '-' * len(content)
    else:
        raise ValueError('Unsupported format, must be .rst or .md')


def version(project_root='.', tag=True):
    """
    Create a new version:
    1. Set new stable version in package_name/__init__.py
    2. Update header in CHANGELOG file, and ask to review CHANGELOG
    3. Create commit for new version, create git tag, and push
    4. Set new development version in package_name/__init__.py, and CHANGELOG
    5. Commit new development version and push
    """
    versioner = Versioner(project_root=project_root)

    current = versioner.current_version()
    release = versioner.release_version()

    release = input_str('Current version in app.yaml is {current}. Enter'
                        ' release version'.format(current=current),
                        default=release)

    if versioner.path_to_changelog:
        versioner.update_changelog_release(release)

        changelog = versioner.path_to_changelog.read_text()

        input_confirm(
            f'\n{versioner.path_to_changelog} content:'
            f'\n\n{changelog}\n',
            abort=True)

    # Replace version number and create tag
    print('Commiting release version: {}'.format(release))
    versioner.commit_version(
        release, msg_template='{package_name} release {new_version}', tag=tag)

    # Create a new dev version and save it
    bumped_version = versioner.bump_up_version()

    print('Creating new section in CHANGELOG...')
    versioner.add_changelog_new_dev_section(bumped_version)
    print('Commiting dev version: {}'.format(bumped_version))
    versioner.commit_version(
        bumped_version,
        msg_template='Bumps up {package_name} to version {new_version}',
        tag=False)

    call(['git', 'push'])
    print('Version {} was created, you are now in {}'.format(
        release, bumped_version))


def upload(tag, production):
    """
    Check outs a tag, uploads to PyPI
    """
    print('Checking out tag {}'.format(tag))
    call(['git', 'checkout', tag])

    current = Versioner().current_version()

    input_confirm('Version in {} tag is {}. Do you want to continue?'.format(
        tag, current),
                  abort=True)

    # create distribution
    call(['rm', '-rf', 'dist/'])
    call(['python', 'setup.py', 'bdist_wheel'])

    print('Publishing to PyPI...')

    if not production:
        call([
            'twine', 'upload', '--repository-url',
            'https://test.pypi.org/legacy/', 'dist/*'
        ])
    else:
        call(['twine', 'upload', 'dist/*'])
